const IExchange = global.IExchange || require('./IExchange')
const UID = global.UID || require('../utils/UID')
const KlineDTO = global.KlineDTO || require('../data').KlineDTO
const AssetDTO = global.AssetDTO || require('../data').AssetDTO
const MarketDTO = global.MarketDTO || require('../data').MarketDTO
const DepthDTO = global.DepthDTO || require('../data').DepthDTO
const BalanceDTO = global.BalanceDTO || require('../data').BalanceDTO
const OrderDTO = global.OrderDTO || require('../data').OrderDTO
const TransactionDTO = global.TransactionDTO || require('../data').TransactionDTO
const KlineDTOArray = global.KlineDTOArray || require('../data').KlineDTOArray
const TransactionDTOArray = global.TransactionDTOArray || require('../data').TransactionDTOArray
const OrderDTOArray = global.OrderDTOArray || require('../data').OrderDTOArray
const BigNumber = global.BigNumber || require('bignumber.js')

/**
  * ccxt CCXTExchange class
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  */
class CCXTExchange extends IExchange {
  constructor ({ exchangeName, protocol, wwwDomain, apiDomain, port, token, apiToken, isMaster, fetchData, options } = {}) {
    const ccxtOptions = {
      token: token || apiToken,
      protocol,
      wwwDomain,
      apiDomain,
      port: port || (options || {}).port || 443,
      verbose: false,
      timeout: 60000
    }
    super(exchangeName, ccxtOptions)
    this.exchangeName = exchangeName
    this.options = options
    this.fetchData = fetchData
    this.isMaster = isMaster
  }

  /**
    * CCXTExchange name getter
    *
    * @author jsb4ch <jsb4ch@jsmex.io>
    * @since 1.0.0
    */
  get name () {
    return this.exchangeName
  }

  /**
  * ccxt CCXTExchange fetch instruments
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns [] MarketDTO
  */
  async loadMarkets () {
    return Promise.resolve()
      .then(() => this.ccxtInstance.loadMarkets())
      .then((result) => {
        if (result) {
          return CCXTExchange.marketToDTOList(result)
        } else {
          const error = { code: 0, message: (`${CCXTExchange.name} fetchInstruments invalid message`) }
          throw error
        }
      })
  }

  /**
  * ccxt CCXTExchange fetch market kline
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @param {String} symbol
  * @returns KlineDTO
  */
  async fetchOHLCV ({ symbol, resolution, since = Date.now(), limit } = {}) {
    const ticker = MarketDTO.formatFromSymbol(symbol, { prefix: '', separator: '/', iso: false })
    return Promise.resolve()
      .then(() => this.ccxtInstance.fetchOHLCV(ticker, resolution, since, limit))
      .then(result => {
        if (result) {
          return CCXTExchange.klineToDTOlist(result, symbol, resolution)
        } else {
          const error = { code: 0, message: (`${CCXTExchange.name} fetchOHLCV invalid message`) }
          throw error
        }
      })
  }

  /**
  * ccxt CCXTExchange fetch depth
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @param {String} symbol
  * @returns {DepthDTO}
  */
  async fetchDepth ({ symbol }, options = {}) {
    const args = [...arguments]
    Object.assign(options, this.options)
    const ticker = MarketDTO.formatFromSymbol(symbol, { prefix: '', separator: '/', iso: false })
    return Promise.resolve()
      .then(() => this.ccxtInstance.fetchOrderBook(ticker))
      .then(result => CCXTExchange.depthToDTO(result, symbol))
      .catch(err => {
        if (err instanceof global.ExchangeNotAvailable || err instanceof global.NetworkError) {
          if (options.retryOn2028) {
            return new Promise((resolve, reject) => {
              setTimeout(() => {
                this.createOrder(...args).then(resolve, reject)
              }, options.retryOn2028)
            })
          } else {
            global.error(err)
            const error = { code: 2028, message: `${CCXTExchange.name} ${err.message}` }
            throw error
          }
        } else {
          global.error(err)
          const error = { code: 2027, message: `${CCXTExchange.name} ${err.message}` }
          throw error
        }
      })
  }

  /**
  * ccxt CCXTExchange get balance method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns BalanceDTO
  */
  async balance () {
    const result = await this.ccxtInstance.fetchBalance()
    return result
    // return CCXTExchange.balanceToDTO(result)
  }

  /**
  * ccxt CCXTExchange put order method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @param {String} symbol
  * @param {String} type
  * @param {String} side
  * @param {String} amount
  * @param {String} price
  * @returns {OrderDTO}
  */
  async createOrder ({ symbol, type, side, amount, price }) {
    const ticker = MarketDTO.formatFromSymbol(symbol, { prefix: '', separator: '/', iso: false })
    const result = await this.ccxtInstance.createOrder(ticker, type, side, amount, price)
    result.symbol = symbol
    return CCXTExchange.orderToDTO(result.info)
  }

  /**
  * ccxt CCXTExchange pipeline method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.61
  * @param {String} symbol
  * @param {String} type
  * @param {Array} pipeline
  * @param {Object} options
  * @returns {[]OrderDTO}
  */
  async pipeline ({ symbol, pipeline = [] } = {}, options = {}) {
    const args = [...arguments]
    Object.assign(options, this.options)
    const ticker = MarketDTO.formatFromSymbol(symbol, { prefix: '', separator: '', iso: false })
    try {
      if (!(pipeline instanceof Array)) {
        throw new Error('pipeline is not instance of array')
      }
      pipeline = pipeline.map(pipeElement => {
        if (!pipeElement.market && !!pipeElement.symbol) {
          pipeElement.symbol = MarketDTO.formatFromSymbol(symbol, { prefix: '', separator: '', iso: false })
        }
        return pipeElement
      })
      if (
        typeof options.splitPipeline === 'number' &&
      pipeline.length > options.splitPipeline
      ) {
        const rounds = []
        let mergedList = []
        do {
          let batch = pipeline.splice(0, options.splitPipeline)
          rounds.push(batch)
        } while (pipeline.length > 0)
        for (const pipe of rounds) {
          const cancelIds = pipe.filter(order => order.type === 'cancel').map(order => order.id)
          let result = await this.ccxtInstance.pipeline(ticker, pipe)
          if (result instanceof Array && result.length) {
            result = result.map(record => {
              if (cancelIds.includes(record.id)) {
                record.isCancel = true
              }
              return record
            })
            mergedList = mergedList.concat(result)
          } else {
            const error = { code: 0, message: `${this.name} invalid pipeline response` }
            throw error
          }
        }
        return CCXTExchange.orderToDTOlist(mergedList, symbol)
      } else {
        const cancelIds = pipeline.filter(order => order.type === 'cancel').map(order => order.id)
        let result = await this.ccxtInstance.pipeline(ticker, pipeline)
        if (result instanceof Array && result.length) {
          result = result.map(record => {
            if (cancelIds.includes(record.id)) {
              record.isCancel = true
            }
            return record
          })
          return CCXTExchange.orderToDTOlist(result, symbol)
        } else {
          const error = { code: 0, message: `${this.name} invalid pipeline response` }
          throw error
        }
      }
    } catch (err) {
      if (err instanceof global.ExchangeNotAvailable || err instanceof global.NetworkError) {
        if (options.retryOn2028) {
          return new Promise((resolve, reject) => {
            setTimeout(() => {
              this.createOrder(...args).then(resolve, reject)
            }, options.retryOn2028)
          })
        } else {
          const error = { code: 2028, message: `${this.name} ${err.message}`, info: err }
          throw error
        }
      } else if (err instanceof global.NetworkError) {
        const error = { code: 2025, message: `${this.name} ${err.message}`, info: err }
        throw error
      } else if (err instanceof global.ExchangeError) {
        const error = { code: 2026, message: `${this.name} ${err.message}`, info: err }
        throw error
      } else {
        const error = { code: 2027, message: `${this.name} ${err.message}`, info: err }
        throw error
      }
    }
  }

  /**
  * ccxt CCXTExchange cancel order method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @param {String} symbol
  * @param {Number} id
  * @returns {OrderDTO}
  */
  async cancelOrder ({ id, symbol }) {
    const ticker = MarketDTO.formatFromSymbol(symbol, { prefix: '', separator: '/', iso: false })
    const result = await this.ccxtInstance.cancelOrder(id, ticker)
    result.symbol = symbol
    return CCXTExchange.orderToDTO(result.info)
  }

  /**
  * ccxt CCXTExchange cancel all order method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns {OrderDTOArray}
  */
  async cancelAllOrders () {
    const result = await this.ccxtInstance.cancelAllOrders()
    return CCXTExchange.orderToDTOlist(result.info)
  }

  /**
  * ccxt CCXTExchange fetchKeyValueStorage method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns {Object}
  */
  async fetchKeyValueStorage () {
    const response = await this.ccxtInstance.fetchKeyValueStorage()
    return response.info.result
  }

  /**
  * ccxt CCXTExchange setKeyValueStorage method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns {Object}
  */
  async setKeyValueStorage ({ key, value, params } = {}) {
    const response = await this.ccxtInstance.setKeyValueStorage(key, value, params)
    return response.info.result
  }

  /**
  * ccxt CCXTExchange setBulkKeyValueStorage method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns {Object}
  */
  async setBulkKeyValueStorage ({ bulk, params } = {}) {
    const response = await this.ccxtInstance.setBulkKeyValueStorage(bulk, params)
    return response.info.result
  }

  /**
  * ccxt CCXTExchange getKeyValueStorage method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns {Object}
  */
  async getKeyValueStorage ({ key, params } = {}) {
    const response = await this.ccxtInstance.getKeyValueStorage(key, params)
    return response.info.result
  }

  /**
  * ccxt CCXTExchange deleteKeyValueStorage method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns {Object}
  */
  async deleteKeyValueStorage ({ key, params } = {}) {
    const response = await this.ccxtInstance.deleteKeyValueStorage(key, params)
    return response.info.result
  }

  /**
  * ccxt CCXTExchange fetch orders method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @param {String} symbol
  * @returns {OrderDTOArray}
  */
  async fetchOrders ({ symbol }) {
    const ticker = MarketDTO.formatFromSymbol(symbol, { prefix: '', separator: '/', iso: false })
    const result = await this.ccxtInstance.fetchOrders(ticker)
    return CCXTExchange.orderToDTOlist(result, symbol)
  }

  /**
  * ccxt CCXTExchange fetch transactions method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @param {String} symbol
  * @returns {[]TransactionDTO}
  */
  async fetchTransactions ({ symbol, since = undefined, limit = undefined, params = {} } = {}) {
    const ticker = MarketDTO.formatFromSymbol(symbol, { prefix: '', separator: '/', iso: false })
    const result = await this.ccxtInstance.fetchTrades(ticker, since, limit, params)
    return CCXTExchange.transactionToDTOlist(result, symbol)
  }

  /**
  * ccxt kraken market object converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @param marketDataLean object
  * @returns MarketDTO
  */
  static marketToDTO (marketDataLean) {
    const market = new MarketDTO(UID.create())

    const stockAsset = new AssetDTO(UID.create())
    stockAsset.alias = marketDataLean.base
    stockAsset.asset = marketDataLean.base

    const moneyAsset = new AssetDTO(UID.create())
    moneyAsset.alias = marketDataLean.quote
    moneyAsset.asset = marketDataLean.quote

    market.exchangeName = CCXTExchange.name
    market.symbolId = `@${CCXTExchange.name}/${stockAsset.alias}-${moneyAsset.alias}`
    market.stock = stockAsset
    market.money = moneyAsset
    market.validate()
    return market
  }

  /**
  * ccxt kraken market list converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @param {Array} marketDataLeanObject
  * @returns {[]MarketDTO}
  */
  static marketToDTOList (marketDataLeanObject) {
    if (!(marketDataLeanObject instanceof Object)) {
      const error = { code: 0, message: 'marketDataLeanList is not instance of Array' }
      throw error
    }
    return Object.keys(marketDataLeanObject).map(
      marketDataLeanKey => {
        const marketDataLean = marketDataLeanObject[marketDataLeanKey]
        return CCXTExchange.marketToDTO(marketDataLean)
      }
    )
  }

  /**
  * ccxt kraken kline object converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns KlineDTO
  */
  static klineToDTO (klineDataLean, symbolId, resolution) {
    const uid = UID.create()
    const kline = new KlineDTO(uid)
    kline.symbolId = symbolId
    kline.resolution = resolution.toString()
    kline.time = klineDataLean[0] / 1000
    kline.open = klineDataLean[1].toString()
    kline.high = klineDataLean[2].toString()
    kline.low = klineDataLean[3].toString()
    kline.close = klineDataLean[4].toString()
    kline.volume = klineDataLean[5].toString()
    kline.validate()
    return kline
  }

  /**
  * ccxt kraken kline object converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns [] KlineDTO
  */
  static klineToDTOlist (klineDataLeanList, symbolId, resolution) {
    if (!(klineDataLeanList instanceof Array)) {
      const error = { code: 0, message: 'klineDataLeanList is not instance of Array' }
      throw error
    }
    let klines = klineDataLeanList
      .map(klineDataLean => CCXTExchange.klineToDTO(klineDataLean, symbolId, resolution))
      .sort((a, b) => b.time - a.time)
    return new KlineDTOArray(...klines)
  }

  /**
  * ccxt CCXTExchange depth object converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns DepthDTO
  */
  static depthToDTO (depthDataLean, symbolId) {
    const uid = UID.create()
    const depth = new DepthDTO(uid)
    depth.symbolId = symbolId
    depth.asks = depthDataLean.asks.map(x => x.splice(0, 2))
    depth.bids = depthDataLean.bids.map(x => x.splice(0, 2))
    depth.validate()
    return depth
  }

  /**
  * ccxt CCXTExchange balance object converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns BalanceDTO
  */
  static balanceToDTO (balanceDataLean) {
    const uid = UID.create()
    const defaultBalance = balanceDataLean.default.balance
    const balance = new BalanceDTO(uid)
    balance.exchangeName = CCXTExchange.name
    balance.balance = defaultBalance
    balance.validate()
    return balance
  }

  /**
  * ccxt CCXTExchange order object converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns OrderDTO
  */
  static orderToDTO (orderDataLean) {
    const uid = UID.create()
    const order = new OrderDTO(uid)

    order.orderId = orderDataLean.id || orderDataLean.orderId
    order.time = orderDataLean.time || Date.now() / 1000
    order.type = orderDataLean.type || 'limit'
    order.side = orderDataLean.side
    order.symbol = orderDataLean.symbol || orderDataLean.market
    order.price = orderDataLean.price ? orderDataLean.price.toString() : undefined
    order.value = (orderDataLean.amount || orderDataLean.value) ? (orderDataLean.amount || orderDataLean.value).toString() : undefined
    order.left = orderDataLean.remaining ? orderDataLean.remaining.toString() : undefined
    order.isCancel = orderDataLean.isCancel
    if (!orderDataLean.isCancel) {
      order.validate()
    }
    return order
  }

  /**
  * ccxt CCXTExchange order list converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns [] OrderDTO
  */
  static orderToDTOlist (orderDataLeanList, symbol) {
    if (!(orderDataLeanList instanceof Array)) {
      const error = 'orderDataLeanList is not instance of Array'
      throw error
    }
    let orders = orderDataLeanList.map(
      orderDataLean => {
        if (symbol) {
          orderDataLean.symbol = symbol
        }
        return CCXTExchange.orderToDTO(orderDataLean.info || orderDataLean)
      }
    )
    return new OrderDTOArray(...orders)
  }

  /**
  * ccxt CCXTExchange transaction object converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns TransactionDTO
  */
  static transactionToDTO (transactionDataLean) {
    const uid = UID.create()
    const transaction = new TransactionDTO(uid)
    transaction.transactionId = transactionDataLean.id || undefined
    transaction.time = transactionDataLean.time
    transaction.type = transactionDataLean.type
    transaction.side = transactionDataLean.side
    transaction.symbol = transactionDataLean.symbol || transactionDataLean.market
    transaction.deal = BigNumber(transactionDataLean.price).times(transactionDataLean.amount).toFixed(8)
    transaction.price = transactionDataLean.price
    transaction.value = transactionDataLean.amount
    transaction.validate()
    return transaction
  }

  /**
  * ccxt CCXTExchange transaction list converter
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.0
  * @returns [] TransactionDTO
  */
  static transactionToDTOlist (transactionDataLeanList, symbol) {
    if (!(transactionDataLeanList instanceof Array)) {
      const error = 'transactionDataLeanList is not instance of Array'
      throw error
    }
    let txs = transactionDataLeanList.map(
      transactionDataLean => {
        transactionDataLean.symbol = symbol
        return CCXTExchange.transactionToDTO(transactionDataLean.info) // info pass original raw
      }
    )
    return new TransactionDTOArray(...txs)
  }
}

module.exports = CCXTExchange
