/**
 * JsmexSymbol classs
 *
 * @author jsb4ch <jsb4ch@jsmex.io>
 * @since 1.0.7
*/
class JsmexSymbol {
  constructor (exchangeName, base, stock) {
    if (
      !base && !stock &&
      typeof exchangeName === 'string'
    ) {
      const symbol = exchangeName
      this._symbol = symbol
    } else if (
      typeof exchangeName === 'string' &&
      typeof base === 'string' &&
      typeof stock === 'string'
    ) {
      this._symbol = `@${exchangeName}/${base}-${stock}`
    }
    if (this.isValid()) {
      throw new Error('Invalid JsmexSymbol')
    }
  }

  /**
   * JsmexSymbol isSymbol method
   *
   * @author jsb4ch <jsb4ch@jsmex.io>
   * @since 1.0.7
   * @return {Boolean}
  */
  isValid () {
    return JsmexSymbol.isSymbol(this._symbol)
  }

  /**
   * JsmexSymbol isSymbol method
   *
   * @author jsb4ch <jsb4ch@jsmex.io>
   * @since 1.0.7
   * @param {String} symbol
   * @return {Boolean}
  */
  static isSymbol (symbol) {
    if (typeof symbol !== 'string') {
      return false
    }
    let matchesFullRegex = symbol.match(JsmexSymbol.fullRegex)
    let matchesSmallRegex = symbol.match(JsmexSymbol.smallRegex)
    if (
      (matchesFullRegex instanceof Array && matchesFullRegex.length === 1) ||
      (matchesSmallRegex instanceof Array && matchesSmallRegex.length === 1)
    ) {
      return true
    } else {
      return false
    }
  }

  /**
   * JsmexSymbol isValid method
   *
   * @author jsb4ch <jsb4ch@jsmex.io>
   * @since 1.0.8
   * @param {String} symbol
   * @return {Boolean}
  */
  static isValid (symbol) {
    return JsmexSymbol.isSymbol(symbol)
  }

  /**
  * JsmexSymbol exchangeNameBySymbol method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.7
  * @param {String} symbol
  * @return {String}
  */
  static exchangeName (symbol) {
    if (!JsmexSymbol.isValid(symbol)) {
      throw new Error('symbol is not string')
    }
    let matches = symbol.match(JsmexSymbol.exchangeNameRegex)
    if (matches instanceof Array) {
      return matches[0].slice(1, -1)
    } else {
      return null
    }
  }

  /**
  * JsmexSymbol getTicker method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.9
  * @param {String} symbol
  * @return {String}
  */
  static getTicker (symbol) {
    if (JsmexSymbol.isSymbol(symbol)) {
      let small = symbol.match(JsmexSymbol.smallRegex)
      if (small) {
        return small[0].replace('/', '-')
      }
      let full = symbol.match(JsmexSymbol.fullRegex)
      if (full) {
        return symbol.split('/').slice(-1)[0]
      }
    }
    return null
  }

  /**
  * JsmexSymbol getTicker method
  *
  * @author jsb4ch <jsb4ch@jsmex.io>
  * @since 1.0.16
  * @param {String} symbol
  * @return {String}
  */
  static formatFromSymbol (symbol, { prefix = '', separator = '' } = {}) {
    let baseTicker = JsmexSymbol.getTicker(symbol)
    if (baseTicker) {
      let splittedBaseTicker = baseTicker.split('-')
      return `${prefix}${splittedBaseTicker[0]}${separator}${splittedBaseTicker[1]}`
    }
    return null
  }

  /**
    * JsmexSymbol full regex get
    *
    * @author jsb4ch <jsb4ch@jsmex.io>
    * @since 1.0.10
    * @return {RegExp}
    */
  static get exchangeNameRegex () {
    return /^@([^]+)\//g
  }
  /**
    * JsmexSymbol full regex get
    *
    * @author jsb4ch <jsb4ch@jsmex.io>
    * @since 1.0.10
    * @return {RegExp}
    */
  static get fullRegex () {
    return /^@([^]+)\/([0-9a-zA-Z]+)-([0-9a-zA-Z]+)$/g
  }

  /**
    * JsmexSymbol small regex get
    *
    * @author jsb4ch <jsb4ch@jsmex.io>
    * @since 1.0.10
    * @return {RegExp}
    */
  static get smallRegex () {
    return /^([0-9a-zA-Z]+)([-/])([0-9a-zA-Z]+)$/g
  }
}

module.exports = JsmexSymbol
